package com.nokia.pm_import.task;

import com.nokia.common.exception.MyRuntimeException;
import com.nokia.common.psql.PsqlUtil;
import com.nokia.common.ssh.SSHUtil;
import com.nokia.pm_import.config.TaskConfig;
import com.xxl.job.core.context.XxlJobHelper;
import com.xxl.job.core.handler.annotation.XxlJob;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVPrinter;
import org.apache.commons.csv.CSVRecord;
import org.slf4j.MDC;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.context.properties.ConfigurationProperties;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * pm 4g小时数据入库定时任务
 */
@ConfigurationProperties("task.pm4g")
@Data
@Component
@Slf4j
public class Pm4gTask {
    /**
     * 无文件次数告警值
     */
    private static final int ALERT_COUNT = 12;
    /**
     * 无文件次数
     */
    private static final AtomicInteger NO_FILE_COUNT = new AtomicInteger(0);
    /**
     * 数据文件目录
     */
    private String sourceDir;
    /**
     * 数据文件下载目录
     */
    private String downloadDir;
    /**
     * 数据文件前缀
     */
    private String filePrefix;
    /**
     * 数据文件去重目录
     */
    private String distinctDir;
    private final DateFormat dateFormat = new SimpleDateFormat("yyyyMMddHH");
    private SSHUtil sshUtil;
    @Autowired
    private TaskConfig taskConfig;

    /**
     * 定时任务
     *
     */
//    @Scheduled(cron = "0 */10 * * * ?")
    @XxlJob("pm4gJobHandler")
    public void pm4gJobHandler() {
        String traceId = UUID.randomUUID().toString().replace("-", "");
        // 日志添加跟踪id
        MDC.put("traceId", traceId);
        try {
            // 创建文件夹
            Files.createDirectories(Paths.get(downloadDir));
            Files.createDirectories(Paths.get(distinctDir));
            sshUtil = new SSHUtil(taskConfig.getSshHost(), taskConfig.getSshPort(), taskConfig.getSshUsername(),
                    taskConfig.getSshPassword());
            // 获取文件列表
            List<String> list = sshUtil.ls(sourceDir)
                    .stream().filter(t -> t.endsWith(".csv")).sorted().collect(Collectors.toList());
            if (CollectionUtils.isEmpty(list)) {
                int count = NO_FILE_COUNT.incrementAndGet();
                log.info("{}次没有文件", count);
                XxlJobHelper.log("{}次没有文件", count);
                // 无文件次数告警
                if (count > ALERT_COUNT) {
                    log.error("超过{}次没有文件", ALERT_COUNT);
                    XxlJobHelper.log("超过{}次没有文件", ALERT_COUNT);
                    // 重置无文件计数
                    NO_FILE_COUNT.set(0);
                }
                return;
            }

            // 重置无文件计数
            NO_FILE_COUNT.set(0);
            log.info("扫描到的文件: {}", list);
            XxlJobHelper.log("扫描到的文件: {}", list);
            for (String t : list) {
                CompletableFuture.runAsync(() -> {
                    try {
                        MDC.put("traceId", traceId);
                        singleTask(t);
                    } catch (InterruptedException e) {
                        log.error("线程中断: {}", e.getMessage(), e);
                        XxlJobHelper.log("线程中断: {}", e.getMessage(), e);
                        Thread.currentThread().interrupt();
                        throw new MyRuntimeException(e);
                    } catch (Exception e) {
                        throw new MyRuntimeException(e);
                    }
                }).get(9, TimeUnit.MINUTES);
            }
        } catch (InterruptedException e) {
            log.error("线程中断: {}", e.getMessage(), e);
            XxlJobHelper.log("线程中断: {}", e.getMessage(), e);
            Thread.currentThread().interrupt();
            XxlJobHelper.handleFail(e.getMessage());
        } catch (TimeoutException e) {
            log.error("执行超时: {}", e.getMessage(), e);
            XxlJobHelper.log("执行超时: {}", e.getMessage(), e);
            XxlJobHelper.handleFail(e.getMessage());
        } catch (Exception e) {
            log.error("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.log("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.handleFail(e.getMessage());
        } finally {
            try {
                sshUtil.disconnect();
            } catch (IOException e) {
                log.error("ssh关闭异常: {}", e.getMessage(), e);
                XxlJobHelper.log("ssh关闭异常: {}", e.getMessage(), e);
                XxlJobHelper.handleFail(e.getMessage());
            }
        }
    }

    /**
     * 将入库失败的已处理数据入库
     */
    @XxlJob("pm4gRestHandler")
    public void pm4gRestHandler() {
        String traceId = UUID.randomUUID().toString().replace("-", "");
        // 日志添加跟踪id
        MDC.put("traceId", traceId);
        // 读取目标文件夹下的文件
        try (Stream<Path> stream = Files.list(Paths.get(distinctDir))) {
            List<Path> pathList = stream.collect(Collectors.toList());
            log.info("pathList: {}", pathList);
            XxlJobHelper.log("pathList: {}", pathList);
            for (Path t : pathList) {
                String filename = t.getFileName().toString();
                log.info("file: {}", filename);
                XxlJobHelper.log("file: {}", filename);
                CompletableFuture.runAsync(() -> {
                    // 日志添加跟踪id
                    MDC.put("traceId", traceId);
                    try {
                        importData(filename);
                    } catch (Exception e) {
                        throw new MyRuntimeException(e);
                    }
                }).get(taskConfig.getTimeout(), TimeUnit.MINUTES);
            }
        } catch (InterruptedException e) {
            log.error("线程中断: {}", e.getMessage(), e);
            XxlJobHelper.log("线程中断: {}", e.getMessage(), e);
            Thread.currentThread().interrupt();
            XxlJobHelper.handleFail(e.getMessage());
        } catch (TimeoutException e) {
            log.error("执行超时: {}", e.getMessage(), e);
            XxlJobHelper.log("执行超时: {}", e.getMessage(), e);
            XxlJobHelper.handleFail(e.getMessage());
        } catch (Exception e) {
            log.error("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.log("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.handleFail(e.getMessage());
        }
    }

    /**
     * 单一任务
     *
     * @param filename 文件名
     */
    public void singleTask(String filename) throws Exception {
        download(filename);
        distinct(filename);
        importData(filename);
    }

    /**
     * 下载文件
     *
     * @param filename 文件名
     */
    public void download(String filename) throws Exception {
        log.info("下载文件: {}", filename);
        XxlJobHelper.log("下载文件: {}", filename);
        String src = sourceDir + filename;
        String dst = downloadDir + filename;
        sshUtil.get(src, dst);
        // 删除远程文件
        log.info("删除远程文件: {}", filename);
        XxlJobHelper.log("删除远程文件: {}", filename);
        sshUtil.rm(src);
    }

    /**
     * 去重
     *
     * @param filename 文件名
     */
    public void distinct(String filename) throws Exception {
        log.info("文件 {} 去重...", filename);
        XxlJobHelper.log("文件 {} 去重...", filename);
        String inputFilePath = downloadDir + filename;
        String outputFilePath = distinctDir + filename;
        Path inputPath = Paths.get(inputFilePath);
        if (Files.size(inputPath) == 0) {
            throw new MyRuntimeException(inputPath.getFileName() + " 空文件");
        }
        try (CSVParser parser = CSVFormat.DEFAULT.builder().build()
                .parse(new InputStreamReader(Files.newInputStream(inputPath), StandardCharsets.UTF_8));
             OutputStreamWriter osw = new OutputStreamWriter(Files.newOutputStream(Paths.get(outputFilePath)),
                     StandardCharsets.UTF_8);
             CSVPrinter printer = new CSVPrinter(osw, CSVFormat.DEFAULT)) {
            Map<String, CSVRecord> map = new HashMap<>();
            // 去重
            for (CSVRecord t : parser) {
                map.put(t.get(2), t);
            }
            if (CollectionUtils.isEmpty(map)) {
                throw new MyRuntimeException(filename + " 数据0条");
            }
            // 添加bom头避免excel乱码
            osw.write('\ufeff');
            // 表头
            printer.printRecord("rpt_time","lacci","cell_name","lte_flow_all","connmax","dn","eci","vendor","sd_name","volte_teletraffic","volte_voice_teletraffic","volte_video_teletraffic","lte_flow_ul","lte_flow_dl","lte_wireless_conn_ratio","succconnestab","attconnestab","nbrsuccestab","nbrattestab","lte_wireless_drop_ratio","attrelenb","attrelenbnormal","succinitalsetup","nbrleft_context","lte_sw_succ_ratio","succoutinterenbs1","succoutinterenbx2","succoutintraenb","attoutinterenbs1","attoutinterenbx2","attoutintraenb","lte_ul_prb_use_ratio","rru_puschprbassn","rru_puschprbtot","lte_dl_prb_use_ratio","rru_pdschprbassn","rru_pdschprbtot","volte_wireless_conn_ratio","volte_voice_conn_ratio","nbrsuccestab_1","nbrsuccestab_2","nbrattestab_1","nbrattestab_2","volte_wireless_drop_ratio","volte_voice_drop_ratio","nbrreqrelenb_1","nbrreqrelenb_normal_1","hofail_1","nbrreqrelenb_2","nbrreqrelenb_normal_2","hofail_2","volte_wireless_drop_ratio_cell","nbrleft_1","nbrhoinc_1","nbrleft_2","nbrhoinc_2","connmean","effectiveconnmean","effectiveconnmax","volte_voice_peak_user_num","activemaxnbrul_1","activemaxnbrdl_1","volte_video_peak_user_num","activemaxnbrul_2","activemaxnbrdl_2","lte_avg_cqi","lte_avg_cqi_fz","lte_avg_cqi_fm","lte_cqi_lt7_ratio","lte_cqi_lt7_ratio_fz","lte_cqi_lt7_ratio_fm","ulmeannl_prb","lte_disturb_highpower_ratio","lte_disturb_highpower_ratio_fz","lte_disturb_highpower_ratio_fm","mr_coverage_ratio","mr_coverage_ratio_fz","mr_coverage_ratio_fm","ue_power_lt0db_ratio","ue_power_lt0db_ratio_fz","ue_power_lt0db_ratio_fm","qci1_ul_avg_loss_ratio","qci1_ul_avg_loss_ratio_fz","qci1_ul_avg_loss_ratio_fm","mr_sinrul_gt0_ratio","mr_sinrul_gt0_ratio_fz","mr_sinrul_gt0_ratio_fm","lte_ul_user_avg_rate","thrptimeul","lte_dl_user_avg_rate","thrptimedl","pdcch_signal_occupy_ratio","rru_pdcchcceutil","rru_pdcchcceavail","lte_flow_per_erab_succ","volte_voice_sw_succ_ratio","succoutinterenbs1_1","succoutinterenbx2_1","succoutintraenb_1","attoutinterenbs1_1","attoutinterenbx2_1","attoutintraenb_1","volte_video_sw_succ_ratio","ho_succoutinterenbs1_2","ho_succoutinterenbx2_2","ho_succoutintraenb_2","ho_attoutinterenbs1_2","ho_attoutinterenbx2_2","ho_attoutintraenb_2","volte_sw_succ_ratio","volte_esrvcc_sw_succ_ratio","succoutgeran","attoutgeran","esrvcc_sw_ratio","qci1_dl_avg_loss_ratio","qci1_dl_avg_loss_ratio_fz","qci1_dl_avg_loss_ratio_fm","qci1_avg_loss_ratio","lte_prb_use_ratio","lte_wireless_drop_ratio_cell","succexecinc","succconnreestab_nonsrccell");
            for (CSVRecord value : map.values()) {
                printer.printRecord(value);
            }
        }

        log.info("文件 {} 去重完成...", filename);
        XxlJobHelper.log("文件 {} 去重完成...", filename);
        // 删除本地源文件
//        Files.deleteIfExists(inputPath);
//        log.info("删除本地源文件 {}...", filename);
//        XxlJobHelper.log("删除本地源文件 {}...", filename);
    }

    /**
     * 导入数据
     *
     * @param filename 文件名
     */
    public void importData(String filename) throws Exception {
        String table = "tsfx.dw_ft_re_st_eutrancell_ind_h";
        String csv = distinctDir + filename;
        String columns = "(rpt_time,lacci,cell_name,lte_flow_all,connmax,dn,eci,vendor,sd_name,volte_teletraffic,volte_voice_teletraffic,volte_video_teletraffic,lte_flow_ul,lte_flow_dl,lte_wireless_conn_ratio,succconnestab,attconnestab,nbrsuccestab,nbrattestab,lte_wireless_drop_ratio,attrelenb,attrelenbnormal,succinitalsetup,nbrleft_context,lte_sw_succ_ratio,succoutinterenbs1,succoutinterenbx2,succoutintraenb,attoutinterenbs1,attoutinterenbx2,attoutintraenb,lte_ul_prb_use_ratio,rru_puschprbassn,rru_puschprbtot,lte_dl_prb_use_ratio,rru_pdschprbassn,rru_pdschprbtot,volte_wireless_conn_ratio,volte_voice_conn_ratio,nbrsuccestab_1,nbrsuccestab_2,nbrattestab_1,nbrattestab_2,volte_wireless_drop_ratio,volte_voice_drop_ratio,nbrreqrelenb_1,nbrreqrelenb_normal_1,hofail_1,nbrreqrelenb_2,nbrreqrelenb_normal_2,hofail_2,volte_wireless_drop_ratio_cell,nbrleft_1,nbrhoinc_1,nbrleft_2,nbrhoinc_2,connmean,effectiveconnmean,effectiveconnmax,volte_voice_peak_user_num,activemaxnbrul_1,activemaxnbrdl_1,volte_video_peak_user_num,activemaxnbrul_2,activemaxnbrdl_2,lte_avg_cqi,lte_avg_cqi_fz,lte_avg_cqi_fm,lte_cqi_lt7_ratio,lte_cqi_lt7_ratio_fz,lte_cqi_lt7_ratio_fm,ulmeannl_prb,lte_disturb_highpower_ratio,lte_disturb_highpower_ratio_fz,lte_disturb_highpower_ratio_fm,mr_coverage_ratio,mr_coverage_ratio_fz,mr_coverage_ratio_fm,ue_power_lt0db_ratio,ue_power_lt0db_ratio_fz,ue_power_lt0db_ratio_fm,qci1_ul_avg_loss_ratio,qci1_ul_avg_loss_ratio_fz,qci1_ul_avg_loss_ratio_fm,mr_sinrul_gt0_ratio,mr_sinrul_gt0_ratio_fz,mr_sinrul_gt0_ratio_fm,lte_ul_user_avg_rate,thrptimeul,lte_dl_user_avg_rate,thrptimedl,pdcch_signal_occupy_ratio,rru_pdcchcceutil,rru_pdcchcceavail,lte_flow_per_erab_succ,volte_voice_sw_succ_ratio,succoutinterenbs1_1,succoutinterenbx2_1,succoutintraenb_1,attoutinterenbs1_1,attoutinterenbx2_1,attoutintraenb_1,volte_video_sw_succ_ratio,ho_succoutinterenbs1_2,ho_succoutinterenbx2_2,ho_succoutintraenb_2,ho_attoutinterenbs1_2,ho_attoutinterenbx2_2,ho_attoutintraenb_2,volte_sw_succ_ratio,volte_esrvcc_sw_succ_ratio,succoutgeran,attoutgeran,esrvcc_sw_ratio,qci1_dl_avg_loss_ratio,qci1_dl_avg_loss_ratio_fz,qci1_dl_avg_loss_ratio_fm,qci1_avg_loss_ratio,lte_prb_use_ratio,lte_wireless_drop_ratio_cell,succexecinc,succconnreestab_nonsrccell)";
        PsqlUtil.copyCsv(taskConfig.getImportScript(), taskConfig.getDbHost1(), taskConfig.getDbPort1(),
                taskConfig.getDbUsername1(), taskConfig.getDbPassword1(), taskConfig.getDbName1(), table, csv,
                columns, taskConfig.getTimeout() - 1, taskConfig.getMinInsertCount());
        PsqlUtil.copyCsv(taskConfig.getImportScript(), taskConfig.getDbHost2(), taskConfig.getDbPort2(),
                taskConfig.getDbUsername2(), taskConfig.getDbPassword2(), taskConfig.getDbName2(), table, csv,
                columns, taskConfig.getTimeout() - 1, taskConfig.getMinInsertCount());
        // 删除重排文件
        Files.deleteIfExists(Paths.get(distinctDir + filename));
    }
}
