package com.nokia.pm_import.task;

import com.nokia.common.exception.MyRuntimeException;
import com.nokia.common.psql.PsqlUtil;
import com.nokia.common.ssh.SSHUtil;
import com.nokia.pm_import.config.TaskConfig;
import com.xxl.job.core.context.XxlJobHelper;
import com.xxl.job.core.handler.annotation.XxlJob;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVPrinter;
import org.apache.commons.csv.CSVRecord;
import org.slf4j.MDC;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.context.properties.ConfigurationProperties;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StringUtils;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * pm 5g小时数据入库定时任务
 */
@ConfigurationProperties("task.pm5g")
@Data
@Component
@Slf4j
public class Pm5gTask {
    /**
     * 无文件次数告警值
     */
    private static final int ALERT_COUNT = 12;
    /**
     * 无文件次数
     */
    private static final AtomicInteger NO_FILE_COUNT = new AtomicInteger(0);
    /**
     * 数据文件目录
     */
    private String sourceDir;
    /**
     * 数据文件下载目录
     */
    private String downloadDir;
    /**
     * 数据文件前缀
     */
    private String filePrefix;
    /**
     * 数据文件去重目录
     */
    private String distinctDir;
    private final DateFormat dateFormat = new SimpleDateFormat("yyyyMMddHH");
    private SSHUtil sshUtil;
    @Autowired
    private TaskConfig taskConfig;

    /**
     * 定时任务
     *
     */
//    @Scheduled(cron = "0 */10 * * * ?")
    @XxlJob("pm5gJobHandler")
    public void pm5gJobHandler() {
        String traceId = UUID.randomUUID().toString().replace("-", "");
        // 日志添加跟踪id
        MDC.put("traceId", traceId);
        try {
            // 创建文件夹
            Files.createDirectories(Paths.get(downloadDir));
            Files.createDirectories(Paths.get(distinctDir));
            sshUtil = new SSHUtil(taskConfig.getSshHost(), taskConfig.getSshPort(), taskConfig.getSshUsername(),
                    taskConfig.getSshPassword());
            // 获取文件列表
            List<String> list = sshUtil.ls(sourceDir)
                    .stream().filter(t -> t.endsWith(".csv")).sorted().collect(Collectors.toList());
            if (CollectionUtils.isEmpty(list)) {
                int count = NO_FILE_COUNT.incrementAndGet();
                log.info("{}次没有文件", count);
                XxlJobHelper.log("{}次没有文件", count);
                // 无文件次数告警
                if (count > ALERT_COUNT) {
                    log.error("超过{}次没有文件", ALERT_COUNT);
                    XxlJobHelper.log("超过{}次没有文件", ALERT_COUNT);
                    // 重置无文件计数
                    NO_FILE_COUNT.set(0);
                }
                return;
            }

            // 重置无文件计数
            NO_FILE_COUNT.set(0);
            log.info("扫描到的文件: {}", list);
            XxlJobHelper.log("扫描到的文件: {}", list);
            for (String t : list) {
                CompletableFuture.runAsync(() -> {
                    try {
                        MDC.put("traceId", traceId);
                        singleTask(t);
                    } catch (InterruptedException e) {
                        log.error("线程中断: {}", e.getMessage(), e);
                        XxlJobHelper.log("线程中断: {}", e.getMessage(), e);
                        Thread.currentThread().interrupt();
                        throw new MyRuntimeException(e);
                    } catch (Exception e) {
                        throw new MyRuntimeException(e);
                    }
                }).get(taskConfig.getTimeout(), TimeUnit.MINUTES);
            }
        } catch (InterruptedException e) {
            log.error("线程中断: {}", e.getMessage(), e);
            XxlJobHelper.log("线程中断: {}", e.getMessage(), e);
            Thread.currentThread().interrupt();
            XxlJobHelper.handleFail(e.getMessage());
        } catch (TimeoutException e) {
            log.error("执行超时: {}", e.getMessage(), e);
            XxlJobHelper.log("执行超时: {}", e.getMessage(), e);
            XxlJobHelper.handleFail(e.getMessage());
        } catch (Exception e) {
            log.error("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.log("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.handleFail(e.getMessage());
        } finally {
            try {
                sshUtil.disconnect();
            } catch (IOException e) {
                log.error("ssh关闭异常: {}", e.getMessage(), e);
                XxlJobHelper.log("ssh关闭异常: {}", e.getMessage(), e);
                XxlJobHelper.handleFail(e.getMessage());
            }
        }
    }

    /**
     * 将入库失败的已处理数据入库
     */
    @XxlJob("pm5gRestHandler")
    public void pm5gRestHandler() {
        String traceId = UUID.randomUUID().toString().replace("-", "");
        // 日志添加跟踪id
        MDC.put("traceId", traceId);
        // 读取目标文件夹下的文件
        try (Stream<Path> stream = Files.list(Paths.get(distinctDir))) {
            List<Path> pathList = stream.collect(Collectors.toList());
            log.info("pathList: {}", pathList);
            XxlJobHelper.log("pathList: {}", pathList);
            for (Path t : pathList) {
                String filename = t.getFileName().toString();
                log.info("file: {}", filename);
                XxlJobHelper.log("file: {}", filename);
                CompletableFuture.runAsync(() -> {
                    // 日志添加跟踪id
                    MDC.put("traceId", traceId);
                    try {
                        importData(filename);
                    } catch (Exception e) {
                        throw new MyRuntimeException(e);
                    }
                }).get(9, TimeUnit.MINUTES);
            }
        } catch (InterruptedException e) {
            log.error("线程中断: {}", e.getMessage(), e);
            XxlJobHelper.log("线程中断: {}", e.getMessage(), e);
            Thread.currentThread().interrupt();
            XxlJobHelper.handleFail(e.getMessage());
        } catch (TimeoutException e) {
            log.error("执行超时: {}", e.getMessage(), e);
            XxlJobHelper.log("执行超时: {}", e.getMessage(), e);
            XxlJobHelper.handleFail(e.getMessage());
        } catch (Exception e) {
            log.error("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.log("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.handleFail(e.getMessage());
        }
    }

    /**
     * 单一任务
     *
     * @param filename 文件名
     */
    public void singleTask(String filename) throws Exception {
        download(filename);
        distinct(filename);
        importData(filename);
    }

    /**
     * 下载文件
     *
     * @param filename 文件名
     */
    public void download(String filename) throws Exception {
        log.info("下载文件: {}", filename);
        XxlJobHelper.log("下载文件: {}", filename);
        String src = sourceDir + filename;
        String dst = downloadDir + filename;
        sshUtil.get(src, dst);
        // 删除远程文件
        log.info("删除远程文件: {}", filename);
        XxlJobHelper.log("删除远程文件: {}", filename);
        sshUtil.rm(src);
    }

    /**
     * 去重
     *
     * @param filename 文件名
     */
    public void distinct(String filename) throws Exception {
        log.info("文件 {} 去重...", filename);
        XxlJobHelper.log("文件 {} 去重...", filename);
        String inputFilePath = downloadDir + filename;
        String outputFilePath = distinctDir + filename;
        Path inputPath = Paths.get(inputFilePath);
        if (Files.size(inputPath) == 0) {
            throw new MyRuntimeException(inputPath.getFileName() + " 空文件");
        }
        try (CSVParser parser = CSVFormat.DEFAULT.builder().build()
                .parse(new InputStreamReader(Files.newInputStream(inputPath), StandardCharsets.UTF_8));
             OutputStreamWriter osw = new OutputStreamWriter(Files.newOutputStream(Paths.get(outputFilePath)),
                     StandardCharsets.UTF_8);
             CSVPrinter printer = new CSVPrinter(osw, CSVFormat.DEFAULT)) {
            Map<String, CSVRecord> map = new HashMap<>();
            // 去重
            for (CSVRecord t : parser) {
                map.put(t.get(2), t);
            }
            if (CollectionUtils.isEmpty(map)) {
                throw new MyRuntimeException(filename + " 数据0条");
            }
            // 添加bom头避免excel乱码
            osw.write('\ufeff');
            // 表头
            printer.printRecord("rpt_time","user_label","nci","enbid","cell_id","enb_cell","vendor","gnb_rrc_conn_ratio","gnb_flow_conn_ratio","gnb_wireless_conn_ratio","gnb_wireless_drop_ratio","gnb_sw_succ_ratio","kpi_rlcupoctul","kpi_rlcupoctdl","gnb_ul_prb_use_ratio","gnb_dl_prb_use_ratio","kpi_pdcchcceoccupancyrate","kpi_ngsig_succconnrate","kpi_flowdroprate_celllevel","kpi_rrcconnreestabrate","kpi_hosuccoutintergnbrate_ng","kpi_hosuccoutintergnbrate_xn","kpi_hosuccoutintergnbrate","kpi_hosuccoutintragnbrate","kpi_hosuccoutrate_intrafreq","kpi_hosuccoutrate_interfreq","kpi_rlcnbrpktlossratedl","kpi_macblerul","kpi_macblerdl","kpi_harqretransrateul","kpi_harqretransratedl","kpi_rank2percentdl","kpi_rank3percentdl","kpi_rank4percentdl","kpi_qpskpercentul","kpi_16qampercentul","kpi_64qampercentul","kpi_256qampercentul","kpi_qpskpercentdl","kpi_16qampercentdl","kpi_64qampercentdl","kpi_256qampercentdl","kpi_dtchprbassnrateul","kpi_dtchprbassnratedl","kpi_avgthrouperprbul","kpi_avgthrouperprbdl","kpi_pdcpupoctul","kpi_pdcpupoctdl","averagelayersul","averagelayersdl","mupairprbrateul","mupairprbratedl","context_attinitalsetup","context_succinitalsetup","context_failinitalsetup","rrc_succconnestab","rrc_attconnestab","flow_nbrsuccestab","flow_nbrattestab","ngsig_connestabsucc","ngsig_connestabatt","context_attrelgnb","context_attrelgnb_normal","context_nbrleft","ho_succexecinc","rrc_succconnreestab_nonsrccell","ho_succoutintercung","ho_succoutintercuxn","ho_succoutintracuinterdu","ho_succoutintradu","ho_attoutintercung","ho_attoutintercuxn","ho_attoutintracuinterdu","ho_attoutcuintradu","rlc_upoctul","rlc_upoctdl","rru_puschprbassn","rru_puschprbtot","rru_pdschprbassn","rru_pdcchcceutil","rru_pdcchcceavail","flow_nbrreqrelgnb","flow_nbrreqrelgnb_normal","flow_hoadmitfail","flow_nbrleft","flow_nbrhoinc","rrc_attconnreestab","ho_succoutintrafreq","ho_attoutexecintrafreq","ho_succoutinterfreq","ho_attoutexecinterfreq","rlc_nbrpktlossdl","rlc_nbrpktdl","mac_nbrreserrtbul","mac_nbrinittbul","mac_nbrreserrtbdl","mac_nbrtbdl","mac_nbrinittbdl","mac_nbrtbdl_rank2","mac_nbrtbdl_rank3","mac_nbrtbdl_rank4","mac_nbrinittbul_qpsk","mac_nbrinittbul_16qam","mac_nbrinittbul_64qam","mac_nbrinittbul_256qam","mac_nbrinittbdl_qpsk","mac_nbrinittbdl_16qam","mac_nbrinittbdl_64qam","mac_nbrinittbdl_256qam","rru_dtchprbassnul","rru_dtchprbassndl","rru_pdschprbtot","mac_cpoctul","mac_cpoctdl","pdcp_upoctul","pdcp_upoctdl","phy_ulmeannl_prb","phy_ulmaxnl_prb","rrc_redirecttolte","rrc_redirecttolte_epsfb","rrc_sannsaconnusermean","rrc_sannsaconnusermax","rrc_connmean","rrc_connmax");
            // 重排
            rerrange(printer, map);
        }

        log.info("文件 {} 去重完成...", filename);
        XxlJobHelper.log("文件 {} 去重完成...", filename);
        // 删除本地源文件
//        Files.deleteIfExists(inputPath);
//        log.info("删除本地源文件 {}...", filename);
//        XxlJobHelper.log("删除本地源文件 {}...", filename);
    }

    /**
     * 导入数据
     *
     * @param filename 文件名
     */
    public void importData(String filename) throws Exception {
        String table = "tsfx.dw_sa_omc_ci_h";
        String csv = distinctDir + filename;
        String columns = "(rpt_time,user_label,nci,enbid,cell_id,enb_cell,vendor,gnb_rrc_conn_ratio,gnb_flow_conn_ratio,gnb_wireless_conn_ratio,gnb_wireless_drop_ratio,gnb_sw_succ_ratio,kpi_rlcupoctul,kpi_rlcupoctdl,gnb_ul_prb_use_ratio,gnb_dl_prb_use_ratio,kpi_pdcchcceoccupancyrate,kpi_ngsig_succconnrate,kpi_flowdroprate_celllevel,kpi_rrcconnreestabrate,kpi_hosuccoutintergnbrate_ng,kpi_hosuccoutintergnbrate_xn,kpi_hosuccoutintergnbrate,kpi_hosuccoutintragnbrate,kpi_hosuccoutrate_intrafreq,kpi_hosuccoutrate_interfreq,kpi_rlcnbrpktlossratedl,kpi_macblerul,kpi_macblerdl,kpi_harqretransrateul,kpi_harqretransratedl,kpi_rank2percentdl,kpi_rank3percentdl,kpi_rank4percentdl,kpi_qpskpercentul,kpi_16qampercentul,kpi_64qampercentul,kpi_256qampercentul,kpi_qpskpercentdl,kpi_16qampercentdl,kpi_64qampercentdl,kpi_256qampercentdl,kpi_dtchprbassnrateul,kpi_dtchprbassnratedl,kpi_avgthrouperprbul,kpi_avgthrouperprbdl,kpi_pdcpupoctul,kpi_pdcpupoctdl,averagelayersul,averagelayersdl,mupairprbrateul,mupairprbratedl,context_attinitalsetup,context_succinitalsetup,context_failinitalsetup,rrc_succconnestab,rrc_attconnestab,flow_nbrsuccestab,flow_nbrattestab,ngsig_connestabsucc,ngsig_connestabatt,context_attrelgnb,context_attrelgnb_normal,context_nbrleft,ho_succexecinc,rrc_succconnreestab_nonsrccell,ho_succoutintercung,ho_succoutintercuxn,ho_succoutintracuinterdu,ho_succoutintradu,ho_attoutintercung,ho_attoutintercuxn,ho_attoutintracuinterdu,ho_attoutcuintradu,rlc_upoctul,rlc_upoctdl,rru_puschprbassn,rru_puschprbtot,rru_pdschprbassn,rru_pdcchcceutil,rru_pdcchcceavail,flow_nbrreqrelgnb,flow_nbrreqrelgnb_normal,flow_hoadmitfail,flow_nbrleft,flow_nbrhoinc,rrc_attconnreestab,ho_succoutintrafreq,ho_attoutexecintrafreq,ho_succoutinterfreq,ho_attoutexecinterfreq,rlc_nbrpktlossdl,rlc_nbrpktdl,mac_nbrreserrtbul,mac_nbrinittbul,mac_nbrreserrtbdl,mac_nbrtbdl,mac_nbrinittbdl,mac_nbrtbdl_rank2,mac_nbrtbdl_rank3,mac_nbrtbdl_rank4,mac_nbrinittbul_qpsk,mac_nbrinittbul_16qam,mac_nbrinittbul_64qam,mac_nbrinittbul_256qam,mac_nbrinittbdl_qpsk,mac_nbrinittbdl_16qam,mac_nbrinittbdl_64qam,mac_nbrinittbdl_256qam,rru_dtchprbassnul,rru_dtchprbassndl,rru_pdschprbtot,mac_cpoctul,mac_cpoctdl,pdcp_upoctul,pdcp_upoctdl,phy_ulmeannl_prb,phy_ulmaxnl_prb,rrc_redirecttolte,rrc_redirecttolte_epsfb,rrc_sannsaconnusermean,rrc_sannsaconnusermax,rrc_connmean,rrc_connmax)";
        PsqlUtil.copyCsv(taskConfig.getImportScript(), taskConfig.getDbHost1(), taskConfig.getDbPort1(),
                taskConfig.getDbUsername1(), taskConfig.getDbPassword1(), taskConfig.getDbName1(), table, csv,
                columns, taskConfig.getTimeout(), taskConfig.getMinInsertCount());
        PsqlUtil.copyCsv(taskConfig.getImportScript(), taskConfig.getDbHost2(), taskConfig.getDbPort2(),
                taskConfig.getDbUsername2(), taskConfig.getDbPassword2(), taskConfig.getDbName2(), table, csv,
                columns, taskConfig.getTimeout(), taskConfig.getMinInsertCount());
        // 删除重排文件
        Files.deleteIfExists(Paths.get(distinctDir + filename));
    }

    /**
     * 保留4位小数
     */
    BigDecimal format(String s) {
        if (!StringUtils.hasText(s)) {
            return null;
        }

        return new BigDecimal(s).setScale(4, RoundingMode.UP);
    }

    /**
     * 重排
     */
    private void rerrange(CSVPrinter printer, Map<String, CSVRecord> map) throws Exception {
        for (CSVRecord t : map.values()) {
            String nciStr = t.get(2);
            String enbid = "";
            String cellId = "";
            // 数据库字段限制100个字符
            String userLabel = org.apache.commons.lang3.StringUtils.substring(t.get(3), 0, 100);
            if (!StringUtils.hasText(nciStr)) {
                continue;
            }

            if (nciStr.contains(".")) {
                String[] a = nciStr.split("\\.");
                enbid = a[1];
                cellId = a[2];
            } else if (nciStr.contains("-")) {
                String[] a = nciStr.split("-");
                enbid = a[1];
                cellId = a[2];
            }

            if (!StringUtils.hasText(enbid) || !StringUtils.hasText(cellId)) {
                continue;
            }
            String enbCell = enbid + "_" + cellId;
            int nci = Integer.parseInt(enbid) * 4096 + Integer.parseInt(cellId);
            printer.printRecord(
                    t.get(1),
                    userLabel,
                    nci,
                    enbid,
                    cellId,
                    enbCell,
                    t.get(6),
                    format(t.get(10)),
                    format(t.get(18)),
                    format(t.get(14)),
                    format(t.get(23)),
                    format(t.get(79)),
                    format(t.get(59)),
                    format(t.get(60)),
                    format(t.get(53)),
                    format(t.get(56)),
                    format(t.get(80)),
                    format(t.get(11)),
                    format(t.get(81)),
                    format(t.get(82)),
                    format(t.get(83)),
                    format(t.get(84)),
                    format(t.get(85)),
                    format(t.get(86)),
                    format(t.get(87)),
                    format(t.get(88)),
                    format(t.get(89)),
                    format(t.get(90)),
                    format(t.get(91)),
                    format(t.get(92)),
                    format(t.get(93)),
                    format(t.get(94)),
                    format(t.get(95)),
                    format(t.get(96)),
                    format(t.get(97)),
                    format(t.get(98)),
                    format(t.get(99)),
                    format(t.get(100)),
                    format(t.get(101)),
                    format(t.get(102)),
                    format(t.get(103)),
                    format(t.get(104)),
                    format(t.get(62)),
                    format(t.get(63)),
                    format(t.get(105)),
                    format(t.get(106)),
                    format(t.get(57)),
                    format(t.get(58)),
                    format(t.get(107)),
                    format(t.get(108)),
                    format(t.get(109)),
                    format(t.get(110)),
                    t.get(111),
                    t.get(112),
                    t.get(113),
                    t.get(8),
                    t.get(7),
                    format(t.get(20)),
                    format(t.get(19)),
                    format(t.get(12)),
                    format(t.get(13)),
                    format(t.get(114)),
                    format(t.get(115)),
                    format(t.get(116)),
                    t.get(117),
                    t.get(118),
                    t.get(119),
                    t.get(120),
                    t.get(121),
                    t.get(122),
                    t.get(123),
                    t.get(124),
                    format(t.get(125)),
                    format(t.get(126)),
                    format(t.get(59)),
                    format(t.get(60)),
                    format(t.get(127)),
                    format(t.get(128)),
                    format(t.get(129)),
                    format(t.get(130)),
                    format(t.get(131)),
                    format(t.get(132)),
                    format(t.get(133)),
                    t.get(134),
                    t.get(135),
                    t.get(136),
                    t.get(137),
                    t.get(138),
                    t.get(139),
                    t.get(140),
                    t.get(141),
                    format(t.get(142)),
                    format(t.get(143)),
                    t.get(144),
                    t.get(145),
                    t.get(147),
                    t.get(148),
                    t.get(149),
                    t.get(150),
                    t.get(151),
                    t.get(152),
                    t.get(153),
                    t.get(154),
                    t.get(155),
                    t.get(156),
                    t.get(157),
                    t.get(158),
                    t.get(159),
                    t.get(160),
                    format(t.get(161)),
                    format(t.get(162)),
                    format(t.get(163)),
                    format(t.get(164)),
                    format(t.get(165)),
                    format(t.get(57)),
                    format(t.get(58)),
                    format(t.get(41)),
                    format(t.get(166)),
                    t.get(167),
                    t.get(168),
                    format(t.get(169)),
                    format(t.get(65)),
                    format(t.get(64)),
                    format(t.get(61))
            );
        }
    }
}
